from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, Literal

import numpy as np
from sklearn.base import TransformerMixin
from sklearn.metrics import DistanceMetric
from sklearn.neighbors import KNeighborsRegressor
from sklearn.utils.validation import _is_arraylike, check_is_fitted

if TYPE_CHECKING:
    from .transformers._base import ComponentReducerMixin


def _validate_data(estimator, *, ensure_all_finite: bool = True, **kwargs):
    """
    Compatibility wrapper around sklearn's _validate_data function.

    scikit-learn >= 1.6.0 moved _validate_data from a method of BaseEstimator to a
    public utility function. This function wraps the utility function if available,
    and falls back to the method if not. `force_all_finite` was also renamed to
    `ensure_all_finite`.

    TODO: Remove when sklearn < 1.6.0 support is dropped.
    """
    try:
        from sklearn.utils.validation import validate_data
    except ImportError:
        return estimator._validate_data(force_all_finite=ensure_all_finite, **kwargs)

    return validate_data(estimator, ensure_all_finite=ensure_all_finite, **kwargs)


class DFIndexCrosswalkMixin:
    """Mixin to crosswalk array indices to dataframe indexes."""

    def _set_dataframe_index_in(self, X):
        """Store dataframe indexes if X is a dataframe."""
        index = getattr(X, "index", None)
        if _is_arraylike(index):
            self.dataframe_index_in_ = np.asarray(index)


class IndependentPredictorMixin:
    """Mixin to return independent predictions based on the X data used
    for fitting the model."""

    def _set_independent_prediction_attributes(self, y):
        """Store independent predictions and score."""
        self.independent_prediction_ = super().predict(X=None)
        self.independent_score_ = super().score(X=None, y=y)


class YFitMixin:
    """Mixin for transformed estimators that use an optional y_fit to fit their
    transformer."""

    def _set_fitted_transformer(self, X, y):
        """Fit and store the transformer, using stored y_fit data if available."""
        y_fit = self.y_fit_ if self.y_fit_ is not None else y
        self.transformer_ = self._get_transformer().fit(X, y_fit)

    def fit(self, X, y, y_fit=None):
        """Fit using transformed feature data. If y_fit is provided, it will be used
        to fit the transformer."""
        self.y_fit_ = y_fit
        return super().fit(X, y)


class RawKNNRegressor(
    DFIndexCrosswalkMixin, IndependentPredictorMixin, KNeighborsRegressor
):
    """
    Subclass of `sklearn.neighbors.KNeighborsRegressor` to support independent
    prediction and scoring and crosswalk array indices to dataframe indexes.
    """

    def fit(self, X, y):
        """Override fit to set attributes using mixins."""
        self._set_dataframe_index_in(X)
        self = super().fit(X, y)
        self._set_independent_prediction_attributes(y)
        return self

    def kneighbors(
        self,
        X=None,
        n_neighbors=None,
        return_distance=True,
        return_dataframe_index=False,
    ):
        """Override kneighbors to optionally return dataframe indexes."""
        neigh_dist, neigh_ind = super().kneighbors(
            X=X, n_neighbors=n_neighbors, return_distance=True
        )

        if return_dataframe_index:
            msg = "Dataframe indexes can only be returned when fitted with a dataframe."
            check_is_fitted(self, "dataframe_index_in_", msg=msg)
            neigh_ind = self.dataframe_index_in_[neigh_ind]

        return (neigh_dist, neigh_ind) if return_distance else neigh_ind


class TransformedKNeighborsRegressor(RawKNNRegressor, ABC):
    """
    Subclass for KNeighbors regressors that apply transformations to the feature data.

    This class serves as a superclass for many estimators in this package, but
    should not be instantiated directly.
    """

    transformer_: TransformerMixin

    @abstractmethod
    def _get_transformer(self) -> TransformerMixin:
        """Return the transformer to use for fitting. Must be implemented by
        subclasses."""
        ...

    def _set_fitted_transformer(self, X, y):
        """Fit and store the transformer."""
        self.transformer_ = self._get_transformer().fit(X, y)

    def fit(self, X, y):
        """Fit using transformed feature data."""
        _validate_data(self, X=X, y=y, ensure_all_finite=True, multi_output=True)
        self._set_dataframe_index_in(X)
        self._set_fitted_transformer(X, y)

        X_transformed = self.transformer_.transform(X)
        return super().fit(X_transformed, y)

    def kneighbors(
        self,
        X=None,
        n_neighbors=None,
        return_distance=True,
        return_dataframe_index=False,
    ):
        """Return neighbor indices and distances using transformed feature data."""
        check_is_fitted(self, "transformer_")
        X_transformed = self.transformer_.transform(X) if X is not None else X
        return super().kneighbors(
            X=X_transformed,
            n_neighbors=n_neighbors,
            return_distance=return_distance,
            return_dataframe_index=return_dataframe_index,
        )


class OrdinationKNeighborsRegressor(TransformedKNeighborsRegressor, ABC):
    """
    Subclass for transformed KNeighbors regressors that apply ordination with
    dimensionality reduction.
    """

    transformer_: ComponentReducerMixin

    def __init__(
        self,
        n_neighbors: int = 5,
        *,
        n_components: int | None = None,
        weights: Literal["uniform", "distance"] | Callable = "uniform",
        algorithm: Literal["auto", "ball_tree", "kd_tree", "brute"] = "auto",
        leaf_size: int = 30,
        p: int = 2,
        metric: str | Callable | DistanceMetric = "minkowski",
        metric_params: dict | None = None,
        n_jobs: int | None = None,
    ):
        super().__init__(
            n_neighbors=n_neighbors,
            weights=weights,
            algorithm=algorithm,
            leaf_size=leaf_size,
            p=p,
            metric=metric,
            metric_params=metric_params,
            n_jobs=n_jobs,
        )
        self.n_components = n_components



import math
from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray


def zero_sum_rows(arr: NDArray) -> NDArray:
    return arr.sum(axis=1) <= 0.0


def zero_sum_columns(arr: NDArray) -> NDArray:
    return arr.sum(axis=0) <= 0.0


# Corresponds to vegan function "initCA" in ordConstrained.R
@dataclass
class InitCA:
    Y_: NDArray
    method: str = "CA"

    @property
    def tot(self):
        return np.sum(self.Y_)

    def _normalized(self):
        return self.Y_ / self.tot

    @property
    def rw(self):
        return np.sum(self._normalized(), axis=1)

    @property
    def cw(self):
        return np.sum(self._normalized(), axis=0)

    @property
    def Y(self):
        rc = np.outer(self.rw, self.cw)
        return (self._normalized() - rc) / np.sqrt(rc)


# Corresponds to vegan function "ordHead" in ordConstrained.R
class OrdHead:
    def __init__(self, transform):
        self.transform = transform

    @property
    def tot_chi(self):
        Y = self.transform.Y
        return np.sum(Y * Y)

    @property
    def Ybar(self):
        return self.transform.Y

    @property
    def grand_total(self):
        return self.transform.tot

    @property
    def rowsum(self):
        return self.transform.rw

    @property
    def colsum(self):
        return self.transform.cw


class ConstrainedOrdination:
    Y_TRANSFORM = {
        "CCA": InitCA,
    }
    ZERO: float = math.sqrt(2.220446e-16)

    @property
    def transform(self):
        return self.Y_TRANSFORM[self.method](self.Y)

    @property
    def head(self):
        return OrdHead(self.transform)

    @property
    def dist_based(self) -> bool:
        return self.transform.method == "DISTBASED"

    @property
    def rw(self):
        return self.transform.rw

    @property
    def cw(self):
        return self.transform.cw

    @property
    def env_center(self):
        return np.average(self.X, axis=0, weights=self.rw)

    @property
    def X_scale(self):
        X = self.X - self.env_center
        return X * np.sqrt(self.rw)[:, np.newaxis]

    @property
    def qr(self):
        return np.linalg.qr(self.X_scale)

    @property
    def Q(self):
        return self.qr[0]

    @property
    def R(self):
        return self.qr[1]

    @property
    def _least_squares(self):
        Q, R = self.qr
        right = Q.T @ self.transform.Y
        return np.linalg.lstsq(R, right, rcond=None)

    @property
    def Y_fit(self):
        return self.X_scale @ self._least_squares[0]

    @property
    def sol(self):
        return np.linalg.svd(self.Y_fit, full_matrices=False)

    @property
    def rank(self):
        ls_rank = self._least_squares[2]
        pos_s = np.sum([self.sol[1] > self.ZERO])
        return min(ls_rank, pos_s)

    @property
    def _U(self):
        return self.sol[0][:, : self.rank]

    @property
    def S(self):
        return np.square(self.sol[1])[: self.rank]

    @property
    def _V(self):
        return self.sol[2].T[:, : self.rank]

    @property
    def U(self):
        return self._U / np.sqrt(self.rw)[:, None]

    @property
    def V(self):
        return self._V / np.sqrt(self.cw)[:, None]

    @property
    def _WA(self):
        x = np.diag(1.0 / np.sqrt(self.S))
        return self.transform.Y.dot(self._V).dot(x)

    @property
    def WA(self):
        return self._WA / np.sqrt(self.rw)[:, None]

    @property
    def biplot_scores(self):
        a = 1.0 / np.sqrt(np.square(self.X_scale).sum(axis=0))
        b = self.X_scale.T.dot(self._U)
        return a[:, None] * b

    @property
    def Y_resid(self):
        return self.transform.Y - self.Y_fit

    @property
    def max_components(self):
        return self.rank

    @property
    def eigenvalues(self):
        return self.S

    @property
    def axis_weights(self):
        return np.diag(np.sqrt(self.S / self.S.sum()))

    @property
    def coefficients(self):
        Q, R = self.qr
        right = Q.T @ self._U
        return np.linalg.lstsq(R, right, rcond=None)[0]

    def projector(self, n_components):
        return (
            self.coefficients[:, :n_components]
            @ self.axis_weights[:n_components, :n_components]
        )

    @property
    def species_scores(self):
        return np.multiply(self.V, np.sqrt(self.S))

    @property
    def site_lc_scores(self):
        return self.U

    @property
    def site_wa_scores(self):
        return self.WA

    @property
    def species_tolerances(self):
        xi = self.site_lc_scores
        uk = self.species_scores
        xiuk = np.zeros((uk.shape[0], xi.shape[0], xi.shape[1]), dtype=np.float64)
        for i, s in enumerate(uk):
            xiuk[i] = xi - s
        y = self.Y.T
        y_xiuk_sqr = np.zeros((uk.shape[0], uk.shape[1]), dtype=np.float64)
        for i in range(y.shape[0]):
            y_xiuk_sqr[i] = y[i] @ np.square(xiuk[i])
        return np.sqrt(y_xiuk_sqr / y.sum(axis=1).reshape(-1, 1))


class CCA(ConstrainedOrdination):
    method = "CCA"
    inertia = "scaled Chi-square"

    # X is environment
    # Y is species
    def __init__(self, X, Y):
        self.X = X
        self.Y = Y
        if np.any(zero_sum_rows(self.Y)):
            msg = "All row sums must be greater than 0"
            raise ValueError(msg)
        self.exclude_Y = zero_sum_columns(self.Y)
        self.Y = self.Y[:, ~self.exclude_Y]
        # Remove excluded species from CCA v and CA v matrices



from ._base import ComponentReducerMixin, StandardScalerWithDOF
from ._cca_transformer import CCATransformer
from ._ccora_transformer import CCorATransformer
from ._mahalanobis_transformer import MahalanobisTransformer

__all__ = [
    "StandardScalerWithDOF",
    "ComponentReducerMixin",
    "CCATransformer",
    "CCorATransformer",
    "MahalanobisTransformer",
]



from sklearn.base import TransformerMixin

from ._base import OrdinationKNeighborsRegressor, YFitMixin
from .transformers import CCATransformer


class GNNRegressor(YFitMixin, OrdinationKNeighborsRegressor):
    """
    Regression using Gradient Nearest Neighbor (GNN) imputation.

    The target is predicted by local interpolation of the targets associated with
    the nearest neighbors in the training set, with distances calculated in transformed
    Canonical Correspondence Analysis (CCA) space.

    See `sklearn.neighbors.KNeighborsRegressor` for more information on parameters
    and implementation.

    Parameters
    ----------
    n_neighbors : int, default=5
        Number of neighbors to use by default for `kneighbors` queries.
    n_components : int, optional
        Number of components to keep during CCA transformation. If `None`, all
        components are kept. If `n_components` is greater than the number of available
        components, an error will be raised.
    weights : {'uniform', 'distance'} or callable, default='uniform'
        Weight function used in prediction.
    algorithm : {'auto', 'ball_tree', 'kd_tree', 'brute'}, default='auto'
        Algorithm used to compute the nearest neighbors.
    leaf_size : int, default=30
        Leaf size passed to `BallTree` or `KDTree`.
    p : int, default=2
        Power parameter for the Minkowski metric.
    metric : str or callable, default='minkowski'
        The distance metric to use for the tree, calculated in CCA space.
    metric_params : dict, default=None
        Additional keyword arguments for the metric function.
    n_jobs : int, optional
        The number of parallel jobs to run for neighbors search. `None` means 1 unless
        in a `joblib.parallel_backend` context. `-1` means using all processors.

    Attributes
    ----------
    effective_metric_ : str or callable
        The distance metric to use. It will be same as the `metric` parameter or a
        synonym of it, e.g. 'euclidean' if the `metric` parameter set to 'minkowski' and
        `p` parameter set to 2.
    effective_metric_params_ : dict
        Additional keyword arguments for the metric function. For most metrics will be
        same with `metric_params` parameter, but may also contain the `p` parameter
        value if the `effective_metric_` attribute is set to 'minkowski'.
    n_features_in_ : int
        Number of features seen during `fit`.
    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during `fit`. Defined only when `X` has feature names
        that are all strings.
    n_samples_fit_ : int
        Number of samples in the fitted data.
    transformer_ : CCATransformer
        Fitted transformer.

    References
    ----------
    Ohmann JL, Gregory MJ. 2002. Predictive Mapping of Forest Composition and Structure
    with Direct Gradient Analysis and Nearest Neighbor Imputation in Coastal Oregon,
    USA. Canadian Journal of Forest Research, 32, 725–741.
    """

    def _get_transformer(self) -> TransformerMixin:
        return CCATransformer(self.n_components)



from sknnr.datasets._base import (
    load_csv_data,
    load_dataset_from_csv_filenames,
    load_moscow_stjoes,
    load_swo_ecoplot,
)

__all__ = [
    "load_csv_data",
    "load_dataset_from_csv_filenames",
    "load_moscow_stjoes",
    "load_swo_ecoplot",
]



from sklearn.base import TransformerMixin

from ._base import TransformedKNeighborsRegressor
from .transformers import MahalanobisTransformer


class MahalanobisKNNRegressor(TransformedKNeighborsRegressor):
    def _get_transformer(self) -> TransformerMixin:
        return MahalanobisTransformer()



import numpy as np
from scipy.stats import f


def ftest_cor(p, q, N, cor):
    """Direct port of yaImpute ftest_cor in yai
    REFERENCE: https://github.com/jeffreyevans/yaImpute/blob/f06d1daf13cf461730c3da5f18f022c1074e9826/R/yai.R#L11-L42
    """  # noqa: E501
    s = min(p, q)
    k = np.arange(s) + 1
    L = np.array([np.prod(1 - np.square(cor[i:s])) for i in range(s)])
    r = (N - s - 1) - ((abs(p - q) + 1) / 2)
    ndf = (p - k + 1) * (q - k + 1)
    u = (ndf - 2) / 4
    xx = (np.square(p - k + 1) + np.square(q - k + 1)) - 5
    t = np.array(
        [
            (
                np.sqrt(((p - k[i] + 1) ** 2 * (q - k[i] + 1) ** 2 - 4) / xx[i])
                if xx[i] > 0
                else 0
            )
            for i in range(s)
        ]
    )
    mask = t > 0
    t = t[mask]
    L = L[mask]
    u = u[mask]
    ndf = ndf[mask]
    L_invt = np.power(L, 1.0 / t)
    ddf = (r * t) - (2 * u)
    set_na = (ddf < 1.0) | (ndf < 1)
    first_na = np.nonzero(set_na)[0]
    first_na = first_na[0] if len(first_na) else len(set_na)
    set_na[first_na:] = True
    F = ((1.0 - L_invt) / L_invt) * (ddf / ndf)
    F[set_na] = np.nan
    pg_F = np.array([1.0 - f.cdf(F[i], ndf[i], ddf[i]) for i in range(F.shape[0])])
    pg_F[set_na] = np.nan
    return pg_F[~np.isnan(pg_F)]


class CCorA:
    """This class closely follows the implementation of
    `statsmodels.multivariate.cancorr.CanCorr` for finding the coefficients associated
    with canonical correlation. The exception to this is the `cscal` property which is
    used to standardize the coefficients and comes from yaImpute.

    We were not able to use the statsmodels implementation directly as it detects
    collinear features and raises a `ValueError` which yaImpute allows to pass silently.
    REFERENCE: https://github.com/statsmodels/statsmodels/blob/77cb066320391ffed4196a32491ddca28e8c9122/statsmodels/multivariate/cancorr.py#L17-L93
    REFERENCE: https://github.com/jeffreyevans/yaImpute/blob/f06d1daf13cf461730c3da5f18f022c1074e9826/R/yai.R#L367-L371
    """  # noqa: E501

    TOLERANCE = 1e-8
    P_VAL = 0.05

    def __init__(self, X, y):
        self.X = np.array(X)
        self.y = np.array(y)
        self.k = min(self.X.shape[1], self.y.shape[1])

    def _apply_svd_with_tolerance(self, arr):
        u, s, v = np.linalg.svd(arr, 0)
        mask = s > self.TOLERANCE
        return u[:, mask], s[mask], v[mask][:, mask]

    @property
    def X_norm(self):
        return self.X - self.X.mean(axis=0)

    @property
    def y_norm(self):
        return self.y - self.y.mean(axis=0)

    @property
    def x_svd(self):
        return self._apply_svd_with_tolerance(self.X_norm)

    @property
    def y_svd(self):
        return self._apply_svd_with_tolerance(self.y_norm)

    @property
    def vx_ds(self):
        _, s, v = self.x_svd
        return v.T / s

    @property
    def vy_ds(self):
        _, s, v = self.y_svd
        return v.T / s

    @property
    def svd(self):
        ux, uy = self.x_svd[0], self.y_svd[0]
        return np.linalg.svd(ux.T.dot(uy), full_matrices=False)

    @property
    def cancorr(self):
        s = self.svd[1]
        return np.array([max(0, min(s[i], 1)) for i in range(len(s))])

    @property
    def cscal(self):
        canonical_coef = self.vx_ds @ self.svd[0][:, 0]
        return 1.0 / np.std(self.X_norm @ canonical_coef, ddof=1)

    @property
    def x_coef(self):
        return self.vx_ds @ self.svd[0][:, : self.k] * self.cscal

    @property
    def y_coef(self):
        return self.vy_ds @ self.svd[2].T[:, : self.k] * self.cscal

    @property
    def f_test(self):
        return ftest_cor(
            self.y_coef.shape[0],
            self.x_coef.shape[0],
            self.y_norm.shape[0],
            self.cancorr,
        )

    @property
    def n_vec(self):
        return max(1, len(self.f_test) - sum(self.f_test > self.P_VAL))

    @property
    def max_components(self):
        return self.n_vec

    def projector(self, n_components):
        return self.x_coef[:, :n_components] @ np.diag(self.cancorr[:n_components])



import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_array, check_is_fitted

from .._base import _validate_data
from . import ComponentReducerMixin, StandardScalerWithDOF
from ._ccora import CCorA


class CCorATransformer(ComponentReducerMixin, TransformerMixin, BaseEstimator):
    def fit(self, X, y):
        _validate_data(self, X=X, reset=True)
        self.scaler_ = StandardScalerWithDOF(ddof=1).fit(X)

        y = check_array(y, input_name="Y", ensure_2d=False, dtype=np.float64)
        if y.ndim == 1:
            y = y.reshape(-1, 1)
        y = StandardScalerWithDOF(ddof=1).fit_transform(y)

        self.ordination_ = CCorA(self.scaler_.transform(X), y)
        self.set_n_components()
        self.projector_ = self.ordination_.projector(n_components=self.n_components_)
        return self

    def transform(self, X, y=None):
        check_is_fitted(self)
        _validate_data(self, X=X, reset=False, ensure_all_finite=True)
        return self.scaler_.transform(X) @ self.projector_

    def fit_transform(self, X, y):
        return self.fit(X, y).transform(X)



from __future__ import annotations

import csv
import types
from dataclasses import dataclass
from importlib import resources
from typing import TYPE_CHECKING

import numpy as np
from numpy.typing import NDArray

if TYPE_CHECKING:  # pragma: no cover
    import pandas as pd

DATA_MODULE = "sknnr.datasets.data"


@dataclass
class Dataset:
    index: NDArray[np.int64] | pd.DataFrame
    data: NDArray[np.float64] | pd.DataFrame
    target: NDArray[np.float64] | pd.DataFrame
    frame: None | pd.DataFrame
    feature_names: list[str]
    target_names: list[str]

    def __repr__(self):
        n = self.data.shape[0]
        n_features = len(self.feature_names)
        n_targets = len(self.target_names)
        return f"Dataset(n={n}, features={n_features}, targets={n_targets})"


def _dataset_as_frame(dataset: Dataset) -> Dataset:
    """Convert a Dataset of arrays to a Dataset of DataFrames."""
    pd = _import_pandas()

    data_df = pd.DataFrame(dataset.data, columns=dataset.feature_names).set_index(
        dataset.index
    )
    target_df = pd.DataFrame(dataset.target, columns=dataset.target_names).set_index(
        dataset.index
    )

    frame = pd.concat([data_df, target_df], axis=1).set_index(dataset.index)

    return Dataset(
        index=dataset.index,
        data=data_df,
        target=target_df,
        frame=frame,
        feature_names=dataset.feature_names,
        target_names=dataset.target_names,
    )


def load_csv_data(
    file_name: str, *, module_name: str | types.ModuleType = DATA_MODULE
) -> tuple[NDArray[np.int64], NDArray[np.float64], NDArray[np.str_]]:
    """Load data from a CSV file from the specified module_name.

    Parameters
    ----------
    file_name: str, required
        The filename of the CSV file to load from `module_name/file_name`.
    module_name: str or module, default='sknnr.datasets.data'
        The module where the data file is located.

    Returns
    -------
    index: ndarray
        The plot IDs from the first column of the CSV file.
    data: ndarray
        The data values from the remaining columns of the CSV file.
    data_names: ndarray
        The column names from the first row of the CSV file.

    Notes
    -----
    The CSV must be formatted with plot IDs in the first column and data values in the
    remaining columns. The first row must contain the column names.
    """
    with resources.files(module_name).joinpath(file_name).open("r") as csv_file:
        data_file = csv.reader(csv_file)
        headers = next(data_file)
        rows = list(iter(data_file))

        index = np.array([row[0] for row in rows], dtype=np.int64)
        data = np.array([row[1:] for row in rows], dtype=np.float64)
        data_names = headers[1:]

    return index, data, data_names


def load_dataset_from_csv_filenames(
    *,
    data_filename: str,
    target_filename: str,
    return_X_y: bool = False,
    as_frame: bool = False,
    module_name: str | types.ModuleType = DATA_MODULE,
) -> tuple[NDArray[np.float64], NDArray[np.float64]] | Dataset:
    """Load separate data and target CSV files into a dataset or paired NumPy arrays.

    Parameters
    ----------
    data_filename: str, required
        The filename of the data CSV file.
    target_filename: str, required
        The filename of the target CSV file.
    return_X_y : bool, default=False
        If True, return the data and target as NumPy arrays instead of a Dataset.
    as_frame : bool, default=False
        If True, the `data` and `target` attributes of the returned Dataset will be
        DataFrames instead of NumPy arrays. The `frame` attribute will also be added as
        a DataFrame with the dataset index. Pandas must be installed for this
        option.
    module_name: str or module, default='sknnr.datasets.data'
        The module where the data files are located.

    Returns
    -------
    Dataset or tuple of ndarray
        A Dataset object containing the data, target, and feature names. If return_X_y
        is True, return a tuple of data and target arrays instead.

    Notes
    -----
    Both CSV files must be formatted with plot IDs in the first column and data values
    in the remaining columns. The first row in each file must contain the column names.
    The plot IDs in each file are expected to match and be in the same order.
    """
    index, data, feature_names = load_csv_data(
        file_name=data_filename, module_name=module_name
    )
    _, target, target_names = load_csv_data(
        file_name=target_filename, module_name=module_name
    )

    dataset = Dataset(
        index=index,
        data=data,
        target=target,
        feature_names=feature_names,
        target_names=target_names,
        frame=None,
    )

    if as_frame:
        dataset = _dataset_as_frame(dataset)

    return (dataset.data, dataset.target) if return_X_y else dataset


def load_moscow_stjoes(
    return_X_y: bool = False, as_frame: bool = False
) -> tuple[NDArray[np.float64], NDArray[np.float64]] | Dataset:
    """Load the Moscow Mountain / St. Joe's dataset (Hudak 2010[^1]).

    The dataset contains 165 plots with environmental, LiDAR, and forest structure
    measurements. Structural measurements of basal area (BA) and tree density (TD)
    are separated by species.

    Parameters
    ----------
    return_X_y : bool, default=False
        If True, return the data and target as NumPy arrays instead of a Dataset.
    as_frame : bool, default=False
        If True, the `data` and `target` attributes of the returned Dataset will be
        DataFrames instead of NumPy arrays. The `frame` attribute will also be added as
        a DataFrame with the dataset index. Pandas must be installed for this
        option.

    Returns
    -------
    Dataset or tuple of ndarray
        A Dataset object containing the data, target, and feature names. If return_X_y
        is True, return a tuple of data and target arrays instead.

    Notes
    -----
    See Hudak 2010[^1] or https://cran.r-project.org/web/packages/yaImpute/yaImpute.pdf
    for more information on the dataset and feature names.

    Reference
    ---------
    [^1] Hudak, A.T. (2010) Field plot measures and predictive maps for "Nearest
    neighbor imputation of species-level, plot-scale forest structure attributes from
    LiDAR data". Fort Collins, CO: U.S. Department of Agriculture, Forest Service,
    Rocky Mountain Research Station.
    https://www.fs.usda.gov/rds/archive/Catalog/RDS-2010-0012
    """
    return load_dataset_from_csv_filenames(
        data_filename="moscow_env.csv",
        target_filename="moscow_spp.csv",
        return_X_y=return_X_y,
        as_frame=as_frame,
    )


def load_swo_ecoplot(
    return_X_y: bool = False, as_frame: bool = False
) -> tuple[NDArray[np.float64], NDArray[np.float64]] | Dataset:
    """Load the southwest Oregon (SWO) USFS Region 6 Ecoplot dataset.

    The dataset contains 3,005 plots with environmental, Landsat, and forest cover
    measurements. Ocular measurements of tree cover (COV) are categorized by
    major tree species present in southwest Oregon.  All data were collected in 2000
    and Landsat imagery processed through the CCDC algorithm was extracted for the
    same year.

    Parameters
    ----------
    return_X_y : bool, default=False
        If True, return the data and target as NumPy arrays instead of a Dataset.
    as_frame : bool, default=False
        If True, the `data` and `target` attributes of the returned Dataset will be
        DataFrames instead of NumPy arrays. The `frame` attribute will also be added as
        a DataFrame with the dataset index. Pandas must be installed for this
        option.

    Returns
    -------
    Dataset or tuple of ndarray
        A Dataset object containing the data, target, and feature names. If return_X_y
        is True, return a tuple of data and target arrays instead.

    Notes
    -----
    These data are a subset of the larger USDA Forest Service Region 6 Ecoplot
    database, which holds 28,000 plots on Region 6 National Forests across Oregon
    and Washington.  The larger database is managed by Patricia Hochhalter (USFS Region
    6 Ecology Program) and used by permission.  Ecoplots were originally used to
    develop plant association guides and are used for a wide array of applications.
    This subset represents plots that were collected in southwest Oregon in 2000.

    Reference
    ---------
    Atzet, T, DE White, LA McCrimmon, PA Martinez, PR Fong, and VD Randall. 1996.
    Field guide to the forested plant associations of southwestern Oregon.
    USDA Forest Service. Pacific Northwest Region, Technical Paper R6-NR-ECOL-TP-17-96.
    """
    return load_dataset_from_csv_filenames(
        data_filename="swo_ecoplot_env.csv",
        target_filename="swo_ecoplot_spp.csv",
        return_X_y=return_X_y,
        as_frame=as_frame,
    )


def _import_pandas():
    """Import pandas and raise an error if it is not installed."""
    try:
        import pandas as pd
    except ImportError:
        msg = (
            "Pandas is required for this functionality. "
            "Please run `pip install pandas` and try again."
        )
        raise ImportError(msg) from None
    return pd



from typing import Union

import numpy as np
from numpy.typing import NDArray
from sklearn.preprocessing import StandardScaler
from sklearn.utils.validation import FLOAT_DTYPES, check_is_fitted

from .._base import _validate_data
from ._cca import CCA
from ._ccora import CCorA

Ordination = Union[CCA, CCorA]


class StandardScalerWithDOF(StandardScaler):
    def __init__(self, ddof=0):
        super().__init__()
        self.ddof = ddof

    def fit(self, X, y=None):
        scaler = super().fit(X, y)

        X = _validate_data(
            self,
            X=X,
            accept_sparse=False,
            dtype=FLOAT_DTYPES,
            ensure_all_finite="allow-nan",
            reset=False,
            ensure_min_samples=self.ddof + 1,
        )
        scaler.scale_ = np.std(X, axis=0, ddof=self.ddof)
        return scaler


class ComponentReducerMixin:
    """
    Mixin for transformers that allow reduction of the number of components.
    """

    ordination_: Ordination

    def __init__(self, n_components=None):
        self.n_components = n_components

    def get_feature_names_out(self) -> NDArray:
        check_is_fitted(self, "n_components_")
        return np.asarray(
            [
                f"{self.ordination_.__class__.__name__.lower()}{i}"
                for i in range(self.n_components_)
            ],
            dtype=object,
        )

    def set_n_components(self):
        n_components = (
            self.n_components
            if self.n_components is not None
            else self.ordination_.max_components
        )
        if not 0 <= n_components <= self.ordination_.max_components:
            raise ValueError(
                f"n_components={n_components} must be between 0 and "
                f"{self.ordination_.max_components}"
            )
        self.n_components_ = n_components



from sklearn.base import TransformerMixin

from ._base import TransformedKNeighborsRegressor
from .transformers import StandardScalerWithDOF


class EuclideanKNNRegressor(TransformedKNeighborsRegressor):
    def _get_transformer(self) -> TransformerMixin:
        return StandardScalerWithDOF(ddof=1)



import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import FLOAT_DTYPES, check_is_fitted

from .._base import _validate_data
from . import ComponentReducerMixin
from ._cca import CCA


class CCATransformer(ComponentReducerMixin, TransformerMixin, BaseEstimator):
    def fit(self, X, y):
        X = _validate_data(
            self,
            X=X,
            reset=True,
            dtype=FLOAT_DTYPES,
            ensure_all_finite=True,
            ensure_min_features=2,
            ensure_min_samples=1,
        )
        y = np.asarray(y)
        if len(y.shape) < 2:
            raise ValueError("`y` must be a 2D array.")

        self.ordination_ = CCA(X, y)
        self.set_n_components()
        self.env_center_ = self.ordination_.env_center
        self.projector_ = self.ordination_.projector(n_components=self.n_components_)
        return self

    def transform(self, X, y=None):
        check_is_fitted(self)
        X = _validate_data(
            self,
            X=X,
            reset=False,
            dtype=FLOAT_DTYPES,
            ensure_all_finite=True,
            ensure_min_features=2,
            ensure_min_samples=1,
        )
        return (X - self.env_center_) @ self.projector_

    def fit_transform(self, X, y):
        return self.fit(X, y).transform(X)

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.target_tags.required = True

        return tags



from .__about__ import __version__  # noqa: F401
from ._base import RawKNNRegressor
from ._euclidean import EuclideanKNNRegressor
from ._gnn import GNNRegressor
from ._mahalanobis import MahalanobisKNNRegressor
from ._msn import MSNRegressor

__all__ = [
    "RawKNNRegressor",
    "EuclideanKNNRegressor",
    "MahalanobisKNNRegressor",
    "MSNRegressor",
    "GNNRegressor",
]



import numpy as np
from sklearn.base import BaseEstimator, OneToOneFeatureMixin, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from .._base import _validate_data
from . import StandardScalerWithDOF


class MahalanobisTransformer(OneToOneFeatureMixin, TransformerMixin, BaseEstimator):
    def fit(self, X, y=None):
        _validate_data(
            self, X=X, ensure_all_finite="allow-nan", reset=True, ensure_min_features=2
        )

        self.scaler_ = StandardScalerWithDOF(ddof=1).fit(X)
        covariance = np.cov(self.scaler_.transform(X), rowvar=False)
        self.transform_ = np.linalg.inv(np.linalg.cholesky(covariance).T)
        return self

    def transform(self, X, y=None):
        check_is_fitted(self)
        _validate_data(self, X=X, ensure_all_finite="allow-nan", reset=False)

        return self.scaler_.transform(X) @ self.transform_

    def fit_transform(self, X, y=None):
        return self.fit(X, y).transform(X)

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.input_tags.allow_nan = True

        return tags
